﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Threading.Tasks;

namespace Platform.Comparison
{
    /// <summary>
    /// A version string comparer
    /// </summary>
    public class VersionStringComparer : IComparer<string>
    {
        /// <summary>
        /// Compare two version strings
        /// </summary>
        /// <param name="version1"></param>
        /// <param name="version2"></param>
        /// <returns></returns>
        public int Compare(string version1, string version2)
        {
            var parts1 = version1.Split('.');
            var parts2 = version2.Split('.');
            for (int i = 0; i < Math.Min(parts1.Length, parts2.Length); i++)
            {
                int value1;
                int value2;
                if (!int.TryParse(parts1[i], out value1)) value1 = 0;
                if (!int.TryParse(parts2[i], out value2)) value2 = 0;
                if (value1 < value2) return -1;
                if (value1 > value2) return 1;
            }

            if (parts1.Length < parts2.Length) return -1;
            if (parts1.Length > parts2.Length) return 1;
            return 0;
        }
    }
}
