﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Linq;
using System.Reflection;
using Platform.ComponentModel;

namespace Platform.Data
{
    public class DataViewModel : ObservableBase, IDataModel, IEditableObject
    {
        public Guid Id
        {
            get => _Id;
            set => SetPropertyValue(ref _Id, value, nameof(Id));
        }
        private Guid _Id;

        public long RowVersion
        {
            get => _RowVersion;
            set => SetPropertyValue(ref _RowVersion, value, nameof(RowVersion));
        }
        private long _RowVersion;

        protected override void OnPropertyChanged(object sender, PropertyChangedEventArgs args)
        {
            base.OnPropertyChanged(sender, args);

            _IsChanged = true;
        }

        #region IEditableObject

        public bool IsChanged
        {
            get => _IsChanged;
        }
        private bool _IsChanged;

        private Dictionary<string, object> _StoredValues;
        private bool _StoredIsChanged;

        public void BeginEdit()
        {
            _StoredIsChanged = _IsChanged;
            var _StoredValues = new Dictionary<string, object>();
            var properties = GetType().GetTypeInfo().DeclaredProperties;
            foreach (var property in properties)
                if (property.CanWrite)
                    _StoredValues.Add(property.Name, property.GetValue(this));
        }

        public void CancelEdit()
        {
            if (_StoredValues == null) return;
            _IsChanged = _StoredIsChanged;

            var properties = GetType().GetTypeInfo().DeclaredProperties;
            foreach (var item in _StoredValues)
            {
                var property = properties.FirstOrDefault(p => p.Name == item.Key);
                if (property != null && property.CanWrite) property.SetValue(this, item.Value);
            }

        }

        public void EndEdit()
        {
            if (_StoredValues != null)
            {
                _StoredValues.Clear();
                _StoredValues = null;
            }
        }

        #endregion
    }
}
