﻿using System;
using System.Collections.Generic;
using System.Text;

namespace Platform.Events
{
    /// <summary>
    /// The class abstracts event handler code
    /// </summary>
    /// <typeparam name="TEventArgs"></typeparam>
    public class EventSubscription<TEventArgs> : IEventSubscription<TEventArgs>
        where TEventArgs : EventArgs
    {
        private static EventHandler<TEventArgs> _GlobalEvents;
        private EventHandler<TEventArgs> _InstanceEvents;

        #region IDisposable Support

        private bool disposedValue = false; // To detect redundant calls

        // This code added to correctly implement the disposable pattern.
        public void Dispose()
        {
            Dispose(true);
        }

        protected virtual void Dispose(bool disposing)
        {
            if (!disposedValue)
            {
                if (disposing)
                {
                    _GlobalEvents -= _InstanceEvents;
                }

                disposedValue = true;
            }
        }

        #endregion

        /// <summary>
        /// Register event handlers
        /// </summary>
        public event EventHandler<TEventArgs> Raised
        {
            add
            {
                _GlobalEvents += value;
                _InstanceEvents += value;
            }
            remove
            {
                _GlobalEvents += value;
                _InstanceEvents += value;
            }
        }

        /// <summary>
        /// Raise the event
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="args"></param>
        public void RaiseEvent(object sender, TEventArgs args)
        {
            _GlobalEvents?.Invoke(sender, args);
        }
    }
}
