﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.IO;

namespace Platform.Extensions
{
    public static class ExceptionExtensions
    {
        /// <summary>
        /// Gets main exception and all inner exceptions
        /// </summary>
        /// <param name="ex"></param>
        /// <returns></returns>
        public static IEnumerable<Exception> GetList(this Exception ex)
        {
            // process all exceptions
            for (; ex != null; ex = ex.InnerException)
                yield return ex;
        }

        /// <summary>
        /// Format entire exception to a multiline printable format
        /// </summary>
        /// <param name="ex"></param>
        /// <returns></returns>
        public static IEnumerable<string> GetText(this Exception ex )
        {
            return GetList(ex).
                SelectMany( (e,i) => e.Message.SplitLines(true).FormatList("{0:00}. ", i+1).
                Concat( e.StackTrace.SplitLines(true).FormatList(" ") ) );
        }
    }
}
