﻿using System;
using System.Linq;
using System.Collections.Generic;
using System.Text;

namespace Platform.Extensions
{
    public static class IEnumerableExtensions
    {
        /// <summary>
        /// Min or Default
        /// </summary>
        /// <typeparam name="T"></typeparam>
        /// <param name="source"></param>
        /// <param name="value"></param>
        /// <returns></returns>
        public static T MinOrDefault<T>(this IEnumerable<T> source, T value)
        {
            if (source.Count() == 0) return value;
            else return source.Min();
        }

        /// <summary>
        /// Min or default
        /// </summary>
        /// <typeparam name="T"></typeparam>
        /// <typeparam name="U"></typeparam>
        /// <param name="source"></param>
        /// <param name="accesser"></param>
        /// <param name="value"></param>
        /// <returns></returns>
        public static U MinOrDefault<T, U>(this IEnumerable<T> source, Func<T, U> accesser, U value)
        {
            if (source.Count() == 0) return value;
            else return source.Min(accesser);
        }

        /// <summary>
        /// Min or Default
        /// </summary>
        /// <typeparam name="T"></typeparam>
        /// <param name="source"></param>
        /// <param name="value"></param>
        /// <returns></returns>
        public static T MaxOrDefault<T>(this IEnumerable<T> source, T value)
        {
            if (source.Count() == 0) return value;
            else return source.Max();
        }

        /// <summary>
        /// Min or default
        /// </summary>
        /// <typeparam name="T"></typeparam>
        /// <typeparam name="U"></typeparam>
        /// <param name="source"></param>
        /// <param name="accesser"></param>
        /// <param name="value"></param>
        /// <returns></returns>
        public static U MaxOrDefault<T, U>(this IEnumerable<T> source, Func<T, U> accesser, U value)
        {
            if (source.Count() == 0) return value;
            else return source.Max(accesser);
        }

        /// <summary>
        /// Process some action on all items
        /// </summary>
        /// <typeparam name="T"></typeparam>
        /// <param name="source"></param>
        /// <param name="action"></param>
        public static void ForEach<T>(this IEnumerable<T> source, Action<T> action)
        {
            foreach (var item in source) action(item);
        }

        /// <summary>
        /// Join distinct values
        /// </summary>
        /// <typeparam name="T"></typeparam>
        /// <param name="source"></param>
        /// <param name="seperator"></param>
        /// <returns></returns>
        public static string JoinDistinct<T>(this IEnumerable<T> source, string seperator)
        {
            if (source == null) return "";
            return string.Join(seperator, source.Select(s=>s.ToString()).Distinct().ToArray());
        }

        /// <summary>
        /// Gets the longest common prefix string
        /// </summary>
        /// <param name="source"></param>
        /// <param name="caseInsensitive"></param>
        /// <returns></returns>
        public static string LongestCommonPrefix(this IEnumerable<string> source, bool caseInsensitive)
        {
            // check for empty collection
            if (source.Count() == 0) return "";

            // check for shortest string
            int length = source.Min(s => s.Length);
            if (length == 0) return "";

            // get first string of minimum length
            string original = source.First().Substring(0,length);

            // compare strings
            if (caseInsensitive )
            {
                original = original.ToUpper();                 
                for (int i = 0; i < length; i++)
                    if (source.Skip(1).Any(s => char.ToUpper(s[i]) != original[i]))
                        return original.Substring(0, i);
            }
            else
            {
                for (int i = 0; i < length; i++)
                    if (source.Skip(1).Any(s => s[i] != original[i]))
                        return original.Substring(0, i);
            }

            // check all others while equal
            // all are equal
            return original;
        }
    }
}
