﻿using System;

#if MF_FRAMEWORK
using Microsoft.SPOT;
#endif

namespace Platform
{
    /// <summary>
    /// Handler for single item events
    /// </summary>
    /// <typeparam name="T"></typeparam>
    /// <param name="sender"></param>
    /// <param name="e"></param>
    public delegate void GenericEventHandler<T>( object sender, GenericEventArgs<T> e );

    /// <summary>
    /// Generic item event arguments
    /// </summary>
    /// <typeparam name="T"></typeparam>
    public class GenericEventArgs<T> : EventArgs
    {
        private T _Item;

        /// <summary>
        /// Default constructor
        /// </summary>
        public GenericEventArgs()
        {
        }

        /// <summary>
        /// Constructs a new ItemEventArgs
        /// </summary>
        /// <param name="value"></param>
        public GenericEventArgs( T value )
        {
            _Item = value;
        }

        /// <summary>
        /// Gets or sets the argument
        /// </summary>
        public T Item
        {
            get { return _Item; }
            set { _Item = value; }
        }

        /// <summary>
        /// Convert to embedded type
        /// </summary>
        /// <param name="source"></param>
        /// <returns></returns>
        public static implicit operator T(GenericEventArgs<T> source)
        {
            if (source == null) return default(T);
            else return source.Item;
        }

        /// <summary>
        /// Convert to embedded type
        /// </summary>
        /// <param name="source"></param>
        /// <returns></returns>
        public static implicit operator GenericEventArgs<T>(T source)
        {
            if (source == null) return null;
            else return new GenericEventArgs<T>(source);
        }
    }
}
