﻿using System;
using System.Text;
using System.Linq;
using System.Collections.Generic;

using Platform.Extensions;

namespace Platform.Net
{
    /// <summary>
    /// A readedr that can read miscellanious 
    /// text based values, mostly network protocol 
    /// related, from a string source
    /// </summary>
    public partial class NetStringReader
    {
        #region Constants

        /// <summary>
        /// Timezone definition
        /// </summary>
        public class TimeZone
        {
            /// <summary>
            /// Construct timezone
            /// </summary>
            /// <param name="name"></param>
            /// <param name="offset"></param>
            public TimeZone( string name, double offset )
            {
                Name = name;
                Offset = offset;
            }

            /// <summary>
            /// Gets the timezone name
            /// </summary>
            public string Name { get; private set; }

            /// <summary>
            /// Gets the timezone offset
            /// </summary>
            public double Offset { get; private set; }
        }

        public static readonly char[] LWS = new char[] { ' ', '\t' };
        public static readonly char[] TOKEN = new char[] { '-', '.', '!', '%', '*', '_', '+', '\'', '`', '~' };
        public static readonly char[] DIGITS = new char[] { '0', '1', '2', '3', '4', '5', '6', '7', '8', '9' };

        public static readonly string[] ShortDayNames = new string[] { "Mon", "Tue", "Wed", "Thu", "Fri", "Sat", "Sun" };
        public static readonly string[] LongDayNames = new string[] { "Monday", "Tuesday", "Wednesday", "Thursday", "Friday", "Saturday", "Sunday" };
        public static readonly string[] ShortMonthNames = new string[] { "Jan", "Feb", "Mar", "Apr", "May", "Jun", "Jul", "Aug", "Sep", "Oct", "Nov", "Dec" };

        public static readonly TimeZone[] TimeZones = new TimeZone[]
        {
            new TimeZone( "A", 1 ),
            new TimeZone( "ACDT", 10.5 ),
            new TimeZone( "ACST", 9.5 ),
            new TimeZone( "ADT", -3 ),
            new TimeZone( "AEDT", 11 ),
            new TimeZone( "AEST", 10 ),
            new TimeZone( "AFT", 4.5 ),
            new TimeZone( "AKDT", -8 ),
            new TimeZone( "AKST", -9 ),
            new TimeZone( "ALMT", 6 ),
            new TimeZone( "ANAST", 12 ),
            new TimeZone( "ANAT", 12 ),
            new TimeZone( "AQTT", 5 ),
            new TimeZone( "ART", -3 ),
            new TimeZone( "AST", -4 ),
            new TimeZone( "AWDT", 9 ),
            new TimeZone( "AWST", 8 ),
            new TimeZone( "AZOST", 0 ),
            new TimeZone( "AZOT", -1 ),
            new TimeZone( "AZST", 5 ),
            new TimeZone( "AZT", 4 ),
            new TimeZone( "B", 2 ),
            new TimeZone( "BNT", 8 ),
            new TimeZone( "BOT", -4 ),
            new TimeZone( "BRST", -2 ),
            new TimeZone( "BRT", -3 ),
            new TimeZone( "BST", 1 ),
            new TimeZone( "BTT", 6 ),
            new TimeZone( "C", 3 ),
            new TimeZone( "CAST", 8 ),
            new TimeZone( "CAT", 2 ),
            new TimeZone( "CCT", 6.5 ),
            new TimeZone( "CDT", -5 ),
            new TimeZone( "CEST", 2 ),
            new TimeZone( "CET", 1 ),
            new TimeZone( "CHADT", 13.75 ),
            new TimeZone( "CHAST", 12.75 ),
            new TimeZone( "CKT", -10 ),
            new TimeZone( "CLST", -3 ),
            new TimeZone( "CLT", -4 ),
            new TimeZone( "COT", -5 ),
            new TimeZone( "CST", -6 ),
            new TimeZone( "CVT", -1 ),
            new TimeZone( "CXT", 7 ),
            new TimeZone( "ChST", 10 ),
            new TimeZone( "D", 4 ),
            new TimeZone( "DAVT", 7 ),
            new TimeZone( "E", 5 ),
            new TimeZone( "EASST", -5 ),
            new TimeZone( "EAST", -6 ),
            new TimeZone( "EAT", 3 ),
            new TimeZone( "ECT", -5 ),
            new TimeZone( "EDT", -4 ),
            new TimeZone( "EEST", 3 ),
            new TimeZone( "EET", 2 ),
            new TimeZone( "EGST", 0 ),
            new TimeZone( "EGT", -1 ),
            new TimeZone( "EST", -5 ),
            new TimeZone( "ET", -5 ),
            new TimeZone( "F", 6 ),
            new TimeZone( "FJST", 13 ),
            new TimeZone( "FJT", 12 ),
            new TimeZone( "FKST", -3 ),
            new TimeZone( "FKT", -4 ),
            new TimeZone( "FNT", -2 ),
            new TimeZone( "G", 7 ),
            new TimeZone( "GALT", -6 ),
            new TimeZone( "GAMT", -9 ),
            new TimeZone( "GET", 4 ),
            new TimeZone( "GFT", -3 ),
            new TimeZone( "GILT", 12 ),
            new TimeZone( "GMT", 0 ),
            new TimeZone( "GST", 4 ),
            new TimeZone( "GYT", -4 ),
            new TimeZone( "H", 8 ),
            new TimeZone( "HAA", -3 ),
            new TimeZone( "HAC", -5 ),
            new TimeZone( "HADT", -9 ),
            new TimeZone( "HAE", -4 ),
            new TimeZone( "HAP", -7 ),
            new TimeZone( "HAR", -6 ),
            new TimeZone( "HAST", -10 ),
            new TimeZone( "HAT", -2.5 ),
            new TimeZone( "HAY", -8 ),
            new TimeZone( "HKT", 8 ),
            new TimeZone( "HLV", -4.5 ),
            new TimeZone( "HNA", -4 ),
            new TimeZone( "HNC", -6 ),
            new TimeZone( "HNE", -5 ),
            new TimeZone( "HNP", -8 ),
            new TimeZone( "HNR", -7 ),
            new TimeZone( "HNT", -3.5 ),
            new TimeZone( "HNY", -9 ),
            new TimeZone( "HOVT", 7 ),
            new TimeZone( "I", 9 ),
            new TimeZone( "ICT", 7 ),
            new TimeZone( "IDT", 3 ),
            new TimeZone( "IOT", 6 ),
            new TimeZone( "IRDT", 4.5 ),
            new TimeZone( "IRKST", 9 ),
            new TimeZone( "IRKT", 9 ),
            new TimeZone( "IRST", 3.5 ),
            new TimeZone( "JST", 9 ),
            new TimeZone( "K", 10 ),
            new TimeZone( "KGT", 6 ),
            new TimeZone( "KRAST", 8 ),
            new TimeZone( "KRAT", 8 ),
            new TimeZone( "KST", 9 ),
            new TimeZone( "KUYT", 4 ),
            new TimeZone( "L", 11 ),
            new TimeZone( "LHDT", 11 ),
            new TimeZone( "LHST", 10.5 ),
            new TimeZone( "LINT", 14 ),
            new TimeZone( "M", 12 ),
            new TimeZone( "MAGST", 12 ),
            new TimeZone( "MAGT", 12 ),
            new TimeZone( "MART", -9.5 ),
            new TimeZone( "MAWT", 5 ),
            new TimeZone( "MDT", -6 ),
            new TimeZone( "MESZ", 2 ),
            new TimeZone( "MEZ", 1 ),
            new TimeZone( "MHT", 12 ),
            new TimeZone( "MMT", 6.5 ),
            new TimeZone( "MSD", 4 ),
            new TimeZone( "MSK", 4 ),
            new TimeZone( "MST", -7 ),
            new TimeZone( "MUT", 4 ),
            new TimeZone( "MVT", 5 ),
            new TimeZone( "MYT", 8 ),
            new TimeZone( "N", -1 ),
            new TimeZone( "NCT", 11 ),
            new TimeZone( "NDT", -2.5 ),
            new TimeZone( "NFT", 11.5 ),
            new TimeZone( "NOVST", 7 ),
            new TimeZone( "NOVT", 6 ),
            new TimeZone( "NPT", 5.75 ),
            new TimeZone( "NST", -3.5 ),
            new TimeZone( "NUT", -11 ),
            new TimeZone( "NZDT", 13 ),
            new TimeZone( "NZST", 12 ),
            new TimeZone( "O", -2 ),
            new TimeZone( "OMSST", 7 ),
            new TimeZone( "OMST", 7 ),
            new TimeZone( "P", -3 ),
            new TimeZone( "PDT", -7 ),
            new TimeZone( "PET", -5 ),
            new TimeZone( "PETST", 12 ),
            new TimeZone( "PETT", 12 ),
            new TimeZone( "PGT", 10 ),
            new TimeZone( "PHOT", 13 ),
            new TimeZone( "PHT", 8 ),
            new TimeZone( "PKT", 5 ),
            new TimeZone( "PMDT", -2 ),
            new TimeZone( "PMST", -3 ),
            new TimeZone( "PONT", 11 ),
            new TimeZone( "PST", -8 ),
            new TimeZone( "PT", -8 ),
            new TimeZone( "PWT", 9 ),
            new TimeZone( "PYST", -3 ),
            new TimeZone( "PYT", -4 ),
            new TimeZone( "Q", -4 ),
            new TimeZone( "R", -5 ),
            new TimeZone( "RET", 4 ),
            new TimeZone( "S", -6 ),
            new TimeZone( "SAMT", 4 ),
            new TimeZone( "SAST", 2 ),
            new TimeZone( "SBT", 11 ),
            new TimeZone( "SCT", 4 ),
            new TimeZone( "SGT", 8 ),
            new TimeZone( "SRT", -3 ),
            new TimeZone( "SST", -11 ),
            new TimeZone( "T", -7 ),
            new TimeZone( "TAHT", -10 ),
            new TimeZone( "TFT", 5 ),
            new TimeZone( "TJT", 5 ),
            new TimeZone( "TKT", 13 ),
            new TimeZone( "TLT", 9 ),
            new TimeZone( "TMT", 5 ),
            new TimeZone( "TVT", 12 ),
            new TimeZone( "U", -8 ),
            new TimeZone( "ULAT", 8 ),
            new TimeZone( "UTC", 0 ),
            new TimeZone( "UYST", -2 ),
            new TimeZone( "UYT", -3 ),
            new TimeZone( "UZT", 5 ),
            new TimeZone( "V", -9 ),
            new TimeZone( "VET", -4.5 ),
            new TimeZone( "VLAST", 11 ),
            new TimeZone( "VLAT", 11 ),
            new TimeZone( "VUT", 11 ),
            new TimeZone( "W", -10 ),
            new TimeZone( "WAST", 2 ),
            new TimeZone( "WAT", 1 ),
            new TimeZone( "WEST", 1 ),
            new TimeZone( "WESZ", 1 ),
            new TimeZone( "WET", 0 ),
            new TimeZone( "WEZ", 0 ),
            new TimeZone( "WFT", 12 ),
            new TimeZone( "WGST", -2 ),
            new TimeZone( "WGT", -3 ),
            new TimeZone( "WIB", 7 ),
            new TimeZone( "WIT", 9 ),
            new TimeZone( "WITA", 8 ),
            new TimeZone( "WT", 0 ),
            new TimeZone( "X", -11 ),
            new TimeZone( "Y", -12 ),
            new TimeZone( "YAKST", 10 ),
            new TimeZone( "YAKT", 10 ),
            new TimeZone( "YAPT", 10 ),
            new TimeZone( "YEKST", 6 ),
            new TimeZone( "YEKT", 6 ),
            new TimeZone( "Z", 0 )
        };

        #endregion

        #region Construction & Fields

        private string _Source;
        private int _Index;

        /// <summary>
        /// Construct new parser
        /// </summary>
        /// <param name="source"></param>
        public NetStringReader( string source )
        {
            _Source = source;
        }

        #endregion

        #region Properties

        /// <summary>
        /// Returns true if end of string
        /// </summary>
        public bool EndOfString
        {
            get { return _Index == _Source.Length; }
        }

        /// <summary>
        /// Gets the source
        /// </summary>
        public string Source
        {
            get { return _Source; }
        }

        #endregion

        #region Tasks Simple Read

        /// <summary>
        /// Read string characters
        /// </summary>
        /// <param name="length"></param>
        /// <returns></returns>
        public string Read(int length)
        {
            length = Math.Min(length, _Source.Length - _Index);
            try { return _Source.Substring(_Index, length); }
            finally { _Index += length; }
        }

        /// <summary>
        /// Read all characters
        /// </summary>
        /// <returns></returns>
        public string ReadAll()
        {
            try { return _Source.Substring(_Index); }
            finally { _Index = _Source.Length; }
        }

        /// <summary>
        /// Skip some characters
        /// </summary>
        /// <param name="count"></param>
        public void Skip(int count)
        {
            _Index = Math.Min(_Source.Length, _Index + count);
        }

        /// <summary>
        /// Peek current character
        /// </summary>
        /// <returns></returns>
        public int Peek()
        {
            if (_Index == _Source.Length) return -1;
            else return _Source[_Index];
        }

        #endregion

        #region Tasks Search

        /// <summary>
        /// Get index if character
        /// </summary>
        /// <param name="c"></param>
        /// <returns></returns>
        public int IndexOf(char c)
        {
            int index = _Source.IndexOf(c, _Index);
            if (index == -1) return -1;
            else return index - _Index;
        }

        /// <summary>
        /// Get index if character
        /// </summary>
        /// <param name="c"></param>
        /// <returns></returns>
        public int IndexOfAny(params char[] terminators)
        {
            int index = _Source.IndexOfAny(terminators, _Index);
            if (index == -1) return -1;
            else return index - _Index;
        }

        #endregion

        #region Tasks Read Patterns

        /// <summary>
        /// Check if starts with any pattern
        /// </summary>
        /// <param name="pattern"></param>
        /// <returns></returns>
        public int StartsWith(params string[] patterns)
        {
            // search all patterns
            for (int index = 0; index < patterns.Length; index++)
            {
                // check for pattern match
                int i = 0;
                var pattern = patterns[index];
                for (int j = _Index; i < pattern.Length && j < _Source.Length; i++, j++)
                    if (pattern[i] != _Source[j])
                        break;

                // check for match
                if (i == pattern.Length)
                    return index;
            }

            // not found
            return -1;
        }

        /// <summary>
        /// Check if starts with any pattern
        /// </summary>
        /// <param name="pattern"></param>
        /// <returns></returns>
        public int StartsWith(bool nocase, params string[] patterns)
        {
            // chek if case
            if (!nocase) 
                return StartsWith(patterns);
            
            // check all patterns
            for (int index = 0; index < patterns.Length; index++)
            {
                // ignore case
                int i = 0;
                var pattern = patterns[index].ToLower();
                for (int j = _Index; i < pattern.Length && j < _Source.Length; i++, j++)
                    if (pattern[i] != char.ToLower(_Source[j]))
                        break;

                // check if found
                if (i == pattern.Length)
                    return index;
            }

            // not found
            return -1;
        }

        /// <summary>
        /// Returns the index of the pattern that 
        /// was read or -1 if a pattern was not read
        /// </summary>
        /// <param name="pattern"></param>
        /// <returns></returns>
        public int ReadPattern(params string[] patterns)
        {
            int index = StartsWith(patterns);
            if (index != -1) _Index += patterns[index].Length;
            return index;
        }

        /// <summary>
        /// Read pattern
        /// </summary>
        /// <param name="pattern"></param>
        /// <returns></returns>
        public int ReadPattern(bool nocase, params string[] patterns)
        {
            int index = StartsWith(nocase, patterns);
            if (index != -1) _Index += patterns[index].Length;
            return index;
        }

        /// <summary>
        /// Check if character can be read
        /// </summary>
        /// <param name="value"></param>
        /// <returns></returns>
        public bool Read(char value)
        {
            if (Peek() != value) return false;
            _Index++;
            return true;
        }

        /// <summary>
        /// Check if string can be read
        /// </summary>
        /// <param name="value"></param>
        /// <returns></returns>
        public string Read(string value)
        {
            int result = ReadPattern(value);
            return result == 0 ? value : "";
        }

        #endregion

        #region Tasks Conditional Read

        /// <summary>
        /// Read terminated string
        /// </summary>
        /// <param name="terminators"></param>
        /// <returns></returns>
        public string ReadIfTerminated(params char[] terminators)
        {
            int index = IndexOfAny(terminators);
            if (index == -1) return "";
            return Read(index);
        }

        /// <summary>
        /// Read all or until
        /// </summary>
        /// <param name="terminators"></param>
        /// <returns></returns>
        public string ReadAllOrUntil(params char[] terminators)
        {
            int index = IndexOfAny(terminators);
            if (index == -1) return ReadAll();
            else return Read(index);
        }

        /// <summary>
        /// Read all or until
        /// </summary>
        /// <param name="terminators"></param>
        /// <returns></returns>
        public string ReadAllOrUntil(char terminator)
        {
            int index = IndexOf(terminator);
            if (index == -1) return ReadAll();
            else return Read(index);
        }

        /// <summary>
        /// Read while characters are valid within set
        /// </summary>
        /// <param name="set"></param>
        /// <returns></returns>
        public string ReadWhile(params char[] set)
        {
            int index = _Index;
            for (; index < _Source.Length && set.Contains(_Source[index]); index++) ;
            try { return _Source.Substring(_Index, index - _Index); }
            finally { _Index = index; }
        }

        /// <summary>
        /// Read while characters are valid within set
        /// </summary>
        /// <param name="set"></param>
        /// <returns></returns>
        public string ReadWhile(Func<char, bool> condition)
        {
            int index = _Index;
            for (; index < _Source.Length && condition(_Source[index]); index++) ;
            try { return _Source.Substring(_Index, index - _Index); }
            finally { _Index = index; }
        }

        /// <summary>
        /// Read while characters are valid within set
        /// </summary>
        /// <param name="set"></param>
        /// <returns></returns>
        public string ReadWhile(int count, Func<char, bool> condition)
        {
            int index = _Index;
            for (; index < _Source.Length && condition(_Source[index]) && count > 0; index++, count--) ;
            try { return _Source.Substring(_Index, index - _Index); }
            finally { _Index = index; }
        }

        /// <summary>
        /// Skip Spaces
        /// </summary>
        public void Skip()
        {
            SkipWhile(LWS);
        }

        /// <summary>
        /// Skip characters in set
        /// </summary>
        /// <param name="source"></param>
        /// <param name="chars"></param>
        /// <param name="start"></param>
        public void SkipWhile(params char[] chars)
        {
            while (_Index < _Source.Length && chars.Contains(_Source[_Index])) _Index++;
        }

        #endregion

        #region Tasks Read Token & Quoted String

        /// <summary>
        /// Returns true if parser starts with TOKEN
        /// </summary>
        /// <returns></returns>
        public bool StartsWithToken()
        {
            for (int j = _Index; j < _Source.Length; j++)
                if (!char.IsLetterOrDigit(_Source[j]) && !TOKEN.Contains(_Source[j])) return false;
                else if (char.IsWhiteSpace(_Source[j])) return true;
            return true;
        }

        /// <summary>
        /// Read a token
        /// </summary>
        /// <param name="source"></param>
        /// <param name="start"></param>
        /// <returns></returns>
        public string ReadToken()
        {
            StringBuilder result = new StringBuilder();
            for (; _Index < _Source.Length && (char.IsLetterOrDigit(_Source[_Index]) || TOKEN.Contains(_Source[_Index]) ); _Index++)
                result.Append(_Source[_Index]);
            return result.ToString();
        }

        /// <summary>
        /// Read all token
        /// </summary>
        /// <param name="source"></param>
        /// <param name="start"></param>
        /// <returns></returns>
        public string ReadTokens()
        {
            StringBuilder result = new StringBuilder();

            while( StartsWithToken() )
            {
                result.Append( ReadToken() );
                result.Append( ReadWhile( ' ' ) );
            }

            return result.ToString().TrimEnd();
        }

        /// <summary>
        /// Read escaped string
        /// </summary>
        /// <returns></returns>
        public string ReadQuotedString()
        {
            return ReadEscapedString('"', '"', '\\', null );
        }

        public delegate char CharacterConvertDelegate(char c);

        /// <summary>
        /// Take quoted string
        /// </summary>
        /// <param name="source"></param>
        /// <param name="start"></param>
        /// <returns></returns>
        public string ReadEscapedString(char start, char stop, char escape, CharacterConvertDelegate escaper)
        {
            // prepare result
            StringBuilder result = new StringBuilder();

            // check for quoted string
            if (_Index < _Source.Length && _Source[_Index] == start) _Index++;
            else return null;

            // parse quoted string
            for (bool isescaped = false; _Index < _Source.Length; _Index++)
            {
                // parse escaped character
                if (isescaped)
                {
                    if (escaper == null) result.Append(_Source[_Index]);
                    else result.Append(escaper(_Source[_Index]));
                    isescaped = false;
                    continue;
                }

                // start escaped character
                if (_Source[_Index] == escape)
                {
                    isescaped = true;
                    continue;
                }

                // end quote
                if (_Source[_Index] == stop)
                {
                    _Index++;
                    return result.ToString();
                }

                // append character
                result.Append(_Source[_Index]);
            }

            // failed to parse quoted string
            return null;
        }

        /// <summary>
        /// Read token or quoted string
        /// </summary>
        /// <param name="Source"></param>
        /// <param name="start"></param>
        /// <returns></returns>
        public string ReadTokenOrQuotedString()
        {
            return Peek() == '\x22' ? ReadQuotedString() : ReadToken();
        }

        /// <summary>
        /// Read token or quoted string
        /// </summary>
        /// <param name="Source"></param>
        /// <param name="start"></param>
        /// <returns></returns>
        public string ReadTokensOrQuotedString()
        {
            return Peek() == '\x22' ? ReadQuotedString() : ReadTokens();
        }

        #endregion

        #region Tasks Read Date & Time

        /// <summary>
        /// Read 24h clock
        /// </summary>
        /// <returns></returns>
        public TimeSpan? ReadTime()
        {
            // store current position
            var index = _Index;

            // parse time
            var hour = Read(2).ToInt32(10, -1).Value;
            if (hour<0 || hour > 23 || Read(1) != ":") goto fail;
            var minute = Read(2).ToInt32(10, -1).Value;
            if (minute < 0 || minute > 59 || Read(1) != ":") goto fail;
            var second = Read(2).ToInt32(10, -1).Value;
            if (second < 0 || second > 59) goto fail;
            return new TimeSpan(hour, minute, second);

        // time failed
        fail:
            _Index = index;
            return null;
        }

        /// <summary>
        /// Read Date: d MMM yy
        /// </summary>
        /// <returns></returns>
        public DateTime? ReadDate1()
        {
            var index = _Index;

            // read date 1
            var day = ReadWhile(2, c => char.IsDigit(c)).ToInt32(10, 0).Value;
            if (day == 0 || Read(1) != " ") goto fail;
            var month = ReadPattern(true, ShortMonthNames) + 1;
            if (month == 0 || Read(1) != " ") goto fail;
            var year = Read(2).ToInt32(10, -1).Value + 1900;
            if (year < 1900 || Read(1) != " ") goto fail;
            return new DateTime(year, month, day, 0, 0, 0, DateTimeKind.Utc);

        fail:
            _Index = index;
            return null;
        }

        /// <summary>
        /// Read Date: d MMM yyyy
        /// </summary>
        /// <returns></returns>
        public DateTime? ReadDate2()
        {
            var index = _Index;

            var day = ReadWhile(2, c => char.IsDigit(c)).ToInt32(10, 0).Value;
            if (day == 0 || Read(1) != " ") goto fail;
            var month = ReadPattern(true, ShortMonthNames) + 1;
            if (month == 0 || Read(1) != " ") goto fail;
            var year = Read(4).ToInt32(10, -1).Value;
            if (year == -1 || Read(1) != " ") goto fail;
            return new DateTime(year, month, day, 0, 0, 0, DateTimeKind.Utc);

        fail:
            _Index = index;
            return null;
        }

        /// <summary>
        /// Read Date: dd-MMM-yy
        /// </summary>
        /// <returns></returns>
        public DateTime? ReadDate3()
        {
            // store index
            var index = _Index;

            // read date
            var day = Read(2).ToInt32(10, 0).Value;
            if (day == 0 || Read(1) != "-") goto fail;
            var month = ReadPattern(true, ShortMonthNames) + 1;
            if (month == 0 || Read(1) != "-") goto fail;
            var year = Read(2).ToInt32(10, -1).Value + 1900;
            if (year < 1900 || Read(1) != " ") goto fail;
            return new DateTime(year, month, day, 0, 0, 0, DateTimeKind.Utc);

        fail:
            _Index = index;
            return null;
        }

        /// <summary>
        /// Read time zone
        /// </summary>
        /// <returns></returns>
        public TimeSpan? ReadZone()
        {
            // store current position
            var index = _Index;

            // get zone
            int zone = StartsWith( TimeZones.Select( z => z.Name ).ToArray() );
            if( zone != -1 ) return TimeSpan.FromHours( TimeZones[zone].Offset );

            // get sign
            var sign = Read(1);
            if (sign != "-" && sign != "+") goto fail;
            var hour = Read(2).ToInt32(10, -1).Value;
            if (hour == -1) goto fail;
            var minute = Read(2).ToInt32(10, -1).Value;
            if (minute == -1) goto fail;

            // handle valid signs
            if (sign == "-") return new TimeSpan(-hour, minute, 0);
            if( sign == "+") return new TimeSpan(hour, minute, 0);

            // check for failure
        fail:
            _Index = index;
            return null;
        }

        /// <summary>
        /// Read RFC 822 date
        /// ddd, d MMM yy HH:mm:ss UT
        /// </summary>
        /// <returns></returns>
        public DateTime? ReadDateTimeRfc822()
        {
            // start reading
            var index = _Index;

            // skip day name, comma and space
            ReadPattern(true, ShortDayNames);
            Read(", ");

            // read date
            var date = ReadDate1();
            if (date == null || Read(1) != " " ) goto fail;

            // read time
            var time = ReadTime();
            if (time == null) goto fail;
            else date = date.Value + time.Value;

            // check for timezone
            TimeSpan? zone = null;
            if( Read(1) == " " ) zone = ReadZone();
            if (zone != null) date = date.Value + zone.Value;            

            // return date
            return date.Value.ToLocalTime();

        fail:
            _Index = index;
            return null;
        }

        /// <summary>
        /// Read RFC 2822 date
        /// ddd, d MMM yyyy HH:mm:ss UT
        /// </summary>
        /// <returns></returns>
        public DateTime? ReadDateTimeRfc2822()
        {
            // start reading
            var index = _Index;

            // skip day name, comma and space
            ReadPattern(true, ShortDayNames);
            Read(", ");

            // read date
            var date = ReadDate1();
            if (date == null || Read(1) != " " ) goto fail;

            // read time
            var time = ReadTime();
            if (time == null) goto fail;
            else date = date.Value + time.Value;

            // check for timezone
            TimeSpan? zone = null;
            if (Read(1) == " ") zone = ReadZone();
            if (zone != null) date = date.Value + zone.Value;

            // return date
            return date.Value.ToLocalTime();

        fail:
            _Index = index;
            return null;
        }

        /// <summary>
        /// Read RFC 850 date
        /// dddd, dd-MMM-yy HH:mm:ss UTC"
        /// </summary>
        /// <param name="value"></param>
        /// <param name="default"></param>
        /// <returns></returns>
        public DateTime? ReadDateTimeRfc850()
        {
            // start reading
            var index = _Index;

            // skip day name, comma and space
            int weekday = ReadPattern(true, LongDayNames);
            if (weekday == -1 || Read(2) != ", ") goto fail;

            // read date
            var date = ReadDate3();
            if (date == null || Read(1) != " " ) goto fail;

            // read time
            var time = ReadTime();
            if (time == null || Read(1) != " " ) goto fail;
            else date = date.Value + time.Value;

            // read zone
            TimeSpan? zone = ReadZone();
            if (zone == null) goto fail;
            else date = date.Value + zone.Value;

            // construct date
            return date.Value.ToLocalTime();

        fail:
            _Index = index;
            return null;
        }

        /// <summary>
        /// Read RFC 1123 date
        /// ddd, d MMM yyyy HH:mm:ss UT
        /// "(" + Wkday + "," + SP + ")?" + Date1 + SP + Time1 + "(" + SP + Zone1 + ")?";
        /// </summary>
        /// <param name="value"></param>
        /// <param name="default"></param>
        /// <returns></returns>
        public DateTime? ReadDateTimeRfc1123()
        {
            // start reading
            var index = _Index;

            // skip day name, comma and space
            ReadPattern(true, ShortDayNames);
            Read(", ");

            // read date
            var date = ReadDate2();
            if (date == null) goto fail;

            // read time
            var time = ReadTime();
            if (time == null || Read(1) != " ") goto fail;
            else date = date.Value + time.Value;

            // check for timezone
            TimeSpan? zone = null;
            if (Read(1) == " ") zone = ReadZone();
            if (zone != null) date = date.Value + zone.Value;

            // construct date
            return date.Value.ToLocalTime();

        fail:
            _Index = index;
            return null;
        }

        /// <summary>
        /// Read ASC date
        /// Wkday + SP + Date3 + SP + Time1 + SP + @"(?<year>\d{4})";
        /// Month + SP + @"(?<day>(\d\d)|(\x20\d))";
        /// </summary>
        /// <param name="value"></param>
        /// <param name="default"></param>
        /// <returns></returns>
        public DateTime? ReadDateTimeAscTime()
        {
            var index = _Index;

            // skip dayname, comma and space
            int weekday = ReadPattern(true, ShortDayNames);
            if (weekday == -1 || Read(1) != " " ) goto fail;

            // get month
            int month = ReadPattern(true, ShortMonthNames) + 1;
            if (month == 0 || Read(1) != " ") goto fail;

            // get day
            int day = Read(2).Trim().ToInt32(10, -1).Value;
            if (day == -1 || Read(1) != " " ) goto fail;

            // get time
            var time = ReadTime();
            if (time == null || Read(1) != " ") goto fail;

            // get year
            var year = Read(4).ToInt32(10, -1).Value;
            if (year == -1) goto fail;

            // get date
            return (new DateTime( year, month, day, 0, 0, 0, DateTimeKind.Utc ) + time.Value).ToLocalTime();

        fail:
            _Index = index;
            return null;
        }

        /// <summary>
        /// Read HTTP date and time
        /// </summary>
        /// <returns></returns>
        public DateTime? ReadDateTimeHttp()
        {
            var date = ReadDateTimeRfc1123();
            if (date == null) date = ReadDateTimeRfc850();
            if (date == null) date = ReadDateTimeAscTime();
            return date;
        }

        #endregion

        #region Tasks Read IP Address

        /// <summary>
        /// Read IP v4 address
        /// </summary>
        /// <param name="address"></param>
        /// <returns></returns>
        public string ReadIp4Address()
        {
            // store index
            var index = _Index;

            // parse parts
            int[] parts = new int[4];

            for( int i=0; i<4; i++ )
            {
                var value = ReadWhile(DIGITS).ToInt32(10, null);
                if ( value == null ) goto fail;
                if( i==3 || ReadPattern(".") == -1 ) goto fail;
                parts[i] = value.Value;
            }

            // get date
            return string.Format("{0}.{1}.{2}.{3}", parts[0], parts[1], parts[2], parts[3]);

        fail:
            _Index = index;
            return null;
        }

        #endregion

#if false

        /// <summary>
        /// Read seperated list of key and value pairs
        /// </summary>
        /// <param name="Source"></param>
        /// <param name="seperator1"></param>
        /// <param name="seperator2"></param>
        /// <param name="start"></param>
        /// <returns></returns>
        public IEnumerable ReadParameterList(char[] itemseperators, char[] keyseperators, bool prefix)
        {
            DictionaryStringString result = new DictionaryStringString( true );

            while (true)
            {
                // check if prefixed list
                if (prefix)
                {
                    Skip();
                    if (EndOfString || !itemseperators.Contains((char)Peek())) return result;
                    else _Index++;
                }

                // read key
                Skip();
                var key = ReadToken();
                if (key == "") return result;

                // skip key/value seperator
                Skip();
                SkipWhile(keyseperators);
                result[key] = ReadTokenOrQuotedString();

                // skip items seperator 
                if (!prefix)
                {
                    Skip();
                    if (EndOfString || !itemseperators.Contains((char)Peek())) return result;
                    else _Index++;
                }
            }
        }

#endif

        #region Tasks Read Line(s)

        /// <summary>
        /// Converts a string 
        /// </summary>
        /// <param name="message"></param>
        /// <returns></returns>
        public string ReadLine(bool ignoreEmptyLines)
        {
            StringBuilder result = new StringBuilder();

            while (_Index < _Source.Length)
            {
                // get character
                char c = _Source[_Index++];

                // ignore LF after CR
                if (c == '\n')
                {
                    if (result.Length > 0 || !ignoreEmptyLines) return result.ToString();
                    result.Length = 0;
                    continue;
                }

                // found CR
                if (c == '\r')
                {
                    if (_Index < _Source.Length && _Source[_Index] == '\n') _Index++;
                    if (result.Length > 0 || !ignoreEmptyLines) return result.ToString();
                    result.Length = 0;
                    continue;
                }

                // append character
                result.Append(c);
            }

            // return result
            if (result.Length > 0) return result.ToString();
            else return null;
        }

        /// <summary>
        /// Read all text lines
        /// </summary>
        /// <param name="ignoreEmptyLines"></param>
        /// <returns></returns>
        public IEnumerable<string> ReadAllLines(bool ignoreEmptyLines)
        {
            while (true)
            {
                var line = ReadLine(ignoreEmptyLines);
                if (line == null) break;
                else yield return line;
            }
        }

        #endregion
    }
}
